import asyncio
import logging
import os
import sys
import traceback
from functools import partial
from typing import Optional, Type

import click
import uvloop
from aio_pyformance.registry import MetricsRegistry, set_global_registry
from pscr_common.log import setup_logging

from query_tool.operations.configuration import OperationConfigurationFile
from query_tool.settings import get_settings, SettingsPseudoBorg, validate_settings
from query_tool.workers.abstract_worker import AbstractWorker
from query_tool.workers.driver_worker import DriverWorker
from query_tool.workers.metrics_worker import MetricsWorker
from query_tool.workers.query_worker import QueryWorker

_logger = logging.getLogger(__name__)


# noinspection PyPep8Naming
async def run_worker(Worker: Type[AbstractWorker], operation_queue: asyncio.Queue, test: bool):
    worker = Worker(operation_queue)
    if test:
        worker.test_mode = True
    try:
        await worker.run()
    except Exception as e:
        _logger.error(f"{type(e)}: {e}\n\n" + "".join(traceback.format_tb(e.__traceback__)))
        sys.exit(1)


def exception_handler(_, ctx):
    _logger.error(ctx["message"])
    if ctx.get("exception"):
        _logger.error(str(ctx["exception"]))
    _logger.error(str(ctx))


@click.command()
@click.option("--config", help="YAML scenario config location, if it doesn't exist one will be created.")
@click.option("--api-hostname", help="API endpoint url")
@click.option("--metrics-server", help="The URL of the influxdb server")
@click.option("--metrics-prefix")
@click.option("--metrics-port")
@click.option("--metrics-type", type=click.Choice(["carbon", "console", "influx"]))
@click.option("--metrics-username")
@click.option("--metrics-password")
@click.option("--query-interval")
@click.option("--query-worker-count")
@click.option("--test", is_flag=True)
@click.option("--env-id", help="The non prod env you want to use: dev/staging, otherwise do not use this option")
def query_tool(
    config: str = None,
    api_hostname: Optional[str] = None,
    metrics_server: Optional[str] = None,
    metrics_prefix: Optional[str] = None,
    metrics_port: Optional[str] = None,
    metrics_type: Optional[str] = None,
    metrics_username: Optional[str] = None,
    metrics_password: Optional[str] = None,
    query_interval: Optional[int] = None,
    query_worker_count: Optional[int] = None,
    test: Optional[bool] = None,
    env_id: Optional[str] = None,
) -> int:
    setup_logging(level=logging.DEBUG if os.getenv("VERBOSE") else logging.INFO)
    logging.getLogger().setLevel(logging.DEBUG if os.getenv("VERBOSE") else logging.INFO)
    logging.getLogger("botocore").setLevel(logging.WARN)
    logging.getLogger("boto3").setLevel(logging.WARN)
    _logger.info("Starting query-tool")
    if api_hostname:
        SettingsPseudoBorg.api_hostname = api_hostname
    if metrics_server:
        SettingsPseudoBorg.metrics_server = metrics_server
    if metrics_prefix:
        SettingsPseudoBorg.metrics_prefix = metrics_prefix
    if metrics_port:
        SettingsPseudoBorg.metrics_port = metrics_port
    if metrics_type:
        SettingsPseudoBorg.metrics_type = metrics_type.lower()
    if metrics_username:
        SettingsPseudoBorg.metrics_username = metrics_username
    if metrics_password:
        SettingsPseudoBorg.metrics_password = metrics_password
    if query_interval:
        SettingsPseudoBorg.query_interval = int(query_interval)
    if query_worker_count:
        SettingsPseudoBorg.query_worker_count = int(query_worker_count)
    if env_id:
        SettingsPseudoBorg.env = env_id
    SettingsPseudoBorg.config_path = config

    validate_settings()

    settings = get_settings()
    _logger.info(f"Settings:\n{settings}")
    regenerate_config = partial(OperationConfigurationFile, config)
    # Just validation
    regenerate_config()

    asyncio.get_event_loop().close()
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(exception_handler)

    set_global_registry(MetricsRegistry())

    operation_queue = asyncio.Queue()
    tasks = [
        asyncio.ensure_future(run_worker(QueryWorker, operation_queue, test))
        for _ in range(int(settings.query_worker_count))
    ]
    tasks.append(asyncio.ensure_future(run_worker(MetricsWorker, operation_queue, test)))
    tasks.append(asyncio.ensure_future(run_worker(DriverWorker, operation_queue, test)))
    loop.run_until_complete(asyncio.wait(tasks))
    loop.close()
    _logger.info("query-tool asyncio loop closed")
    return 0
