import logging
import os
import sys
from typing import AnyStr

import pkg_resources
import yaml

import query_tool
from query_tool.operations.operation import OperationFactory, Mode, OperationType
from query_tool.settings import get_settings

_logger = logging.getLogger(__name__)


class OperationMisconfigurationException(BaseException):
    pass


class OperationConfigurationFile:
    def __init__(self, configuration_file):
        self.filepath = configuration_file
        self.check_file()
        with open(self.filepath, "r", encoding="utf8") as stream:
            self.raw = yaml.safe_load(stream)
        self.validate()
        self.operations = [
            OperationConfig(**op_cfg)
            for op_cfg in self.raw["operation_config"]
            if tuple(op_cfg.items())[0][1]["mode"] == get_settings().execution_mode
        ]

    def check_file(self):
        exists = os.path.exists(self.filepath)
        is_dir = os.path.isdir(self.filepath)
        if exists and is_dir:
            raise OperationMisconfigurationException(f"{self.filepath} must be a file, not a directory.")
        if not exists:
            with open(self.filepath, "w+", encoding="utf8") as stream_o:
                raw_str = pkg_resources.resource_string(
                    query_tool.__name__, os.path.join("resources", "default_config.yml")
                )
                stream_o.write(raw_str.decode("utf-8"))
                sys.exit(5)

    def validate(self):
        if not self.raw:
            raise OperationMisconfigurationException("Config is null or empty.")
        if not self.raw.get("operation_config"):
            raise OperationMisconfigurationException('No defined top-level "operation_config:"')
        for operation in self.raw["operation_config"]:
            name, parameters = next(iter(operation.items()))
            if "operation" not in parameters.keys():
                raise OperationMisconfigurationException(
                    f"Operation {name} does not have a set operation "
                    f'(one of {", ".join(OperationType.__members__)}).'
                )
            if "mode" not in parameters.keys():
                raise OperationMisconfigurationException(
                    f"Operation {name} does not have a set mode " f'(one of {", ".join(Mode.__members__)}).'
                )


class OperationConfig:
    def __init__(self, **kwargs):
        """
        :param kwargs: This is the raw object under the 'operation_config' key in the yaml config file.
                       Each element in that 'operation_config' yaml list is one of these objects.
        """
        if len(kwargs.keys()) > 1:
            raise RuntimeError("Only pass one operation_config at a time.")
        self.id, params = next(iter(kwargs.items()))
        self.mode = Mode[params["mode"]]
        self.op_type = OperationType[params["operation"]]
        self.operation = OperationFactory.generate(self.op_type)
        params.pop("mode")
        params.pop("operation")
        self.params = params

    async def execute(self, org_id: AnyStr, q_position: int):
        _logger.debug(f"Executing operation {self.id} ({self.op_type}, {self.mode})")
        await self.operation.execute(org_id, q_position, **self.params)


class MacroOperation:
    """
    This is a wrapper object which is fed through the queue from a single DriverWorker to multiple QueryWorkers
    """

    def __init__(self, operation: OperationConfig, org_id: AnyStr, q_position: int):
        self.operation = operation
        self.org_id = org_id
        self.q_position = q_position

    async def execute(self):
        await self.operation.execute(self.org_id, self.q_position)

    def __repr__(self):
        return f"{self.operation.id} [{self.org_id}]"
