import logging
import traceback
from asyncio import sleep
from enum import Enum
from time import time
from typing import Dict, Optional

from aio_pyformance.registry import global_registry
from aiohttp import ClientError
from dateparser import parse as dateparse
from pscr_common.stream import stream

from query_tool.query import QueryFactory
from query_tool.settings import get_full_prefix, get_settings

_logger = logging.getLogger(__name__)


class Mode(Enum):
    performance = 1
    availability = 2
    availability_summary = 3
    availability_details = 4


class OperationType(Enum):
    process_v2 = 1
    enriched = 2
    summary = 3
    process_details_v2 = 4
    enriched_details = 5


def _replace_time_range(delta: str) -> str:
    return dateparse(delta).strftime("%Y-%m-%dT%H:%M:%SZ")


class AbstractOperation:
    def __init__(self):
        self.registry = global_registry()

    async def _start(self, org_id: str, **kwargs) -> Dict:
        raise NotImplementedError()

    async def _results(self, org_id: str, job_id: str) -> Dict:
        raise NotImplementedError()

    async def _retrieve(self, org_id: str, job_id: str) -> Optional[Dict]:
        rsp = await self._results(org_id, job_id)
        if "contacted" not in rsp or "completed" not in rsp:
            _logger.error(f"Malformed get response for {org_id}/{job_id}:\n{rsp}")
        if rsp["contacted"] == rsp["completed"]:
            return rsp
        meta = rsp
        meta.pop("results", None)
        _logger.debug(f"Search {org_id}/{job_id} incomplete:\n{meta}")

    async def _handle_failed_start(self, rsp: Dict, q_position: int):
        if "translation_key" in rsp and "no_searchers_error" in rsp["translation_key"]:
            await self.registry.meter(f"{get_full_prefix(q_position)}.no_data").mark()
        else:
            await self.registry.meter(f"{get_full_prefix(q_position)}.error_start").mark()

    # noinspection PyBroadException
    async def execute(self, org_id: str, q_position: int, **kwargs) -> Optional[Dict]:
        """
        :param org_id:
        :param q_position:
        :param kwargs: Raw content of config with the mode and operation fields stripped out
        :return: the last result call or None if something went wrong
        """
        send_perf_metrics = True
        if kwargs.get("no_metrics"):
            kwargs.pop("no_metrics")
            send_perf_metrics = False

        if kwargs.get("time_range"):
            if kwargs.get("time_range").get("start"):
                kwargs["time_range"]["start"] = _replace_time_range(kwargs["time_range"]["start"])
            if kwargs.get("time_range").get("start"):
                kwargs["time_range"]["end"] = _replace_time_range(kwargs["time_range"]["end"])

        timeout_threshold = get_settings().timeout_seconds
        start = time()
        _logger.debug(f"Starting query: {org_id} -> {kwargs}")
        try:
            rsp = await self._start(org_id, **kwargs)
            if "job_id" not in rsp:
                _logger.error(f"Malformed start response for {org_id}: {rsp}")
                await self._handle_failed_start(rsp, q_position)
                return
            job_id = rsp["job_id"]
            _logger.debug(f"Successfully started query {org_id}/{job_id}")
        except (ClientError, Exception):
            _logger.error(traceback.format_exc())
            await self.registry.meter(f"{get_full_prefix(q_position)}.error_start").mark()
            return

        completed = False
        timed_out = False
        op_duration = None
        get_count = 0
        while not completed:
            await sleep(get_settings().results_delay)
            try:
                rsp = await self._retrieve(org_id, job_id)
                get_count += 1
                completed = bool(rsp)
            except (ClientError, Exception):
                _logger.error(traceback.format_exc())
                await self.registry.meter(f"{get_full_prefix(q_position)}.error_get").mark()
            op_duration = time() - start
            if op_duration > timeout_threshold:
                timed_out = True
                break

        _logger.info(
            f"Finished query {org_id}/{job_id}. GET Count: {get_count} "
            f"Duration: {op_duration} Timed out: {timed_out}"
        )

        if send_perf_metrics:
            await self._publish_metrics(op_duration, timed_out, q_position, get_count)
        return rsp

    async def _publish_metrics(self, op_duration: float, timed_out: bool, q_position: int, get_count: int):
        prefix = get_full_prefix(q_position)
        duration_h = self.registry.histogram(f"{prefix}.query_duration")
        duration_g = self.registry.gauge(f"{prefix}.query_duration_gauge")
        timeout_m = self.registry.meter(f"{prefix}.query_timeout")
        blocked_m = self.registry.meter(f"{prefix}.query_blocked")
        total_m = self.registry.meter(f"{prefix}.query_total")
        get_g = self.registry.gauge(f"{prefix}.get_count")
        if timed_out and get_count <= 15:
            await blocked_m.mark()
        if timed_out:
            await timeout_m.mark()
        await total_m.mark()
        await duration_h.add(op_duration)
        duration_g.set_value(op_duration)
        get_g.set_value(get_count)


# pylint is wrongly assuming this is a concrete class when it's just an abstract subclass of an abstract parent
# pylint: disable=W0223
class AbstractDetailsOperation(AbstractOperation):
    def __init__(self, operation_type, search_job_payload_key, entry_identifier):
        self.OperationType = operation_type
        self.search_job_payload_key = search_job_payload_key
        self.entry_identifier = entry_identifier
        super().__init__()

    async def execute(self, org_id: str, q_position: int, **kwargs) -> Optional[Dict]:
        rsp = await self.OperationType().execute(org_id, q_position, **kwargs, no_metrics=True)
        if not rsp:
            _logger.error(
                f"Something went wrong, cannot start {self.__class__.__name__} search job for org_id: {org_id}"
            )
            return
        entry_set = rsp["results"]
        _logger.debug(f"Acquired {len(entry_set)} results for {kwargs.get('query')}")
        entry = stream(entry_set).first_or_else()
        if not entry:
            _logger.error(f"No valid {self.__class__.__name__} in dataset for org_id {org_id}")
            return
        # noinspection PyUnresolvedReferences
        args = {self.search_job_payload_key: [entry[self.entry_identifier]]}
        _logger.debug(f"Starting {self.__class__.__name__} query [{org_id}] {args}")
        rsp = await super().execute(org_id, q_position, **args)
        _logger.debug(f"{self.__class__.__name__} result: {rsp}")
        return rsp


class ProcessOperation(AbstractOperation):
    async def _start(self, org_id: str, **kwargs) -> Dict:
        start = QueryFactory.get_from_string(org_id, "start")
        return await start(**kwargs)

    async def _results(self, org_id: str, job_id: str) -> Dict:
        results = QueryFactory.get_from_string(org_id, "results")
        return await results(job_id)


class ProcessDetailsOperation(AbstractDetailsOperation):
    def __init__(self):
        super().__init__(ProcessOperation, "process_guids", "process_guid")

    async def _start(self, org_id: str, **kwargs) -> Dict:
        start = QueryFactory.get_from_string(org_id, "start_details")
        return await start(**kwargs)

    async def _results(self, org_id: str, job_id: str) -> Dict:
        results = QueryFactory.get_from_string(org_id, "results_details")
        return await results(job_id)


class EnrichedOperation(AbstractOperation):
    async def _start(self, org_id: str, **kwargs) -> Dict:
        start = QueryFactory.get_from_string(org_id, "start_enriched")
        return await start(**kwargs)

    async def _results(self, org_id: str, job_id: str) -> Dict:
        results = QueryFactory.get_from_string(org_id, "results_enriched")
        return await results(job_id)


class EnrichedDetailsOperation(AbstractDetailsOperation):
    def __init__(self):
        super().__init__(EnrichedOperation, "event_ids", "event_id")

    async def _start(self, org_id: str, **kwargs) -> Dict:
        start = QueryFactory.get_from_string(org_id, "start_enriched_details")
        return await start(**kwargs)

    async def _results(self, org_id: str, job_id: str) -> Dict:
        results = QueryFactory.get_from_string(org_id, "results_enriched_details")
        return await results(job_id)


class SummaryOperation(AbstractOperation):
    async def _start(self, org_id: str, **kwargs) -> Dict:
        start = QueryFactory.get_from_string(org_id, "start_summary")
        return await start(**kwargs)

    async def _results(self, org_id: str, job_id: str) -> Dict:
        results = QueryFactory.get_from_string(org_id, "results_summary")
        return await results(job_id)

    async def execute(self, org_id: str, q_position: int, **kwargs) -> Optional[Dict]:
        op_config = self.SummaryOperationConfig(**kwargs)
        rsp = await ProcessOperation().execute(
            org_id, q_position, query=op_config.query, rows=op_config.rows, no_metrics=True
        )
        if not rsp:
            _logger.error(f"Something went wrong, cannot start summary search for org_id: {org_id}")
            return
        process_set = rsp["results"]
        _logger.debug(f"Acquired {len(rsp['results'])} results for {op_config.query}:use_parent={op_config.use_parent}")
        if op_config.use_parent:
            process = stream(process_set).filter(lambda e: "parent_guid" in e and e["parent_guid"]).first_or_else()
        else:
            process = stream(process_set).first_or_else()
        if not process:
            _logger.error(
                f"No valid process in dataset for org_id {org_id} (use_parent: {op_config.use_parent}), perhaps only "
                f"enriched events? "
            )
            return
        # noinspection PyUnresolvedReferences
        args = {"process_guid": process["process_guid"], "parent_guid": process.get("parent_guid")}
        _logger.debug(f"Starting summary query [{org_id}] {args}")
        rsp = await super().execute(org_id, q_position, **args)
        _logger.debug(f"Summary result: {rsp}")
        return rsp

    class SummaryOperationConfig:
        def __init__(self, query: str, use_parent: bool = False, rows: int = 100):
            self.query = query
            self.use_parent = use_parent
            self.rows = rows


class OperationFactory:
    @staticmethod
    def generate(op_type: OperationType) -> AbstractOperation:
        if op_type == OperationType.process_v2:
            return ProcessOperation()
        if op_type == OperationType.process_details_v2:
            return ProcessDetailsOperation()
        if op_type == OperationType.enriched:
            return EnrichedOperation()
        if op_type == OperationType.enriched_details:
            return EnrichedDetailsOperation()
        if op_type == OperationType.summary:
            return SummaryOperation()
