import logging
from typing import AnyStr, Dict, List, Optional, Sequence
from asyncio import TimeoutError
from aiohttp import ClientError, ClientSession, ClientTimeout
from query_tool.settings import get_settings

_logger = logging.getLogger(__name__)


class InvalidQueryError(BaseException):
    pass


class AsyncSearchAPI:
    def __init__(self, org_id: AnyStr):
        self.org_id = org_id
        self.url_prefix = f"http://{get_settings().api_hostname}:5003/internal"

    async def _api_request(self, route: AnyStr, method: AnyStr = "GET", params: Optional[Dict] = None) -> Dict:
        def _log_bad_api(rsp, e):
            _logger.error(f"Api error for {method} {self.url_prefix + route}:\n{rsp}\n{e}")

        # Timeout is 60s just like gateway
        async with ClientSession(timeout=ClientTimeout(total=60.0)) as session:
            if method == "GET":
                async with session.get(self.url_prefix + route) as response:
                    try:
                        return await response.json()
                    except (ClientError, TimeoutError) as e:
                        _log_bad_api(response, e)
                        raise
            elif method == "POST":
                async with session.post(self.url_prefix + route, json=params) as response:
                    try:
                        return await response.json()
                    except (ClientError, TimeoutError) as e:
                        _log_bad_api(response, e)
                        raise
            else:
                raise InvalidQueryError(f'Invalid method "{method}" ({route})')

    @staticmethod
    def _fill_params(
        query: str,
        criteria: Optional[Dict[str, Sequence[str]]] = None,
        exclusions: Optional[Dict[str, Sequence[str]]] = None,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        start: Optional[int] = None,
        time_range: Optional[Dict[str, str]] = None,
        use_sqs: Optional[bool] = None,
    ) -> Dict:
        params = {"query": query}
        if criteria is not None:
            params.update(criteria=criteria)
        if exclusions is not None:
            params.update(exclusions=exclusions)
        if fields is not None:
            params.update(fields=fields)
        if rows is not None:
            params.update(rows=rows)
        if sort is not None:
            params.update(sort=sort)
        if start is not None:
            params.update(start=start)
        if time_range is not None:
            params.update(time_range=time_range)
        if use_sqs is not None:
            params.update({"cb.use_sqs": True})
        return params

    async def create_process_result_v2(
        self,
        query: str,
        criteria: Optional[Dict[str, Sequence[str]]] = None,
        exclusions: Optional[Dict[str, Sequence[str]]] = None,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        start: Optional[int] = None,
        time_range: Optional[Dict[str, str]] = None,
        use_sqs: Optional[bool] = None,
    ) -> Dict:
        """
        Creates a process search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param query: a lucene query like "process_name:svchost.exe"
        :param criteria: in the format {"device_name": ["Win7x64"]}
        :param exclusions: in the format {"device_name": ["Win7x64"]}
        :param fields: list of fields in the format ["process_username"]
        :param rows: integer
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param start: integer
        :param time_range: in the format {"start": "2020-01-18T18:34:04Z", "end": "2020-01-18T18:34:04Z"}
        :param use_sqs: represents "cb.use_sqs" to force sqs path for a search
        :return: an ApiResponse with either a valid response or error information
        """
        params = self._fill_params(
            query=query,
            criteria=criteria,
            exclusions=exclusions,
            fields=fields,
            rows=rows,
            sort=sort,
            start=start,
            time_range=time_range,
            use_sqs=use_sqs,
        )
        return await self._api_request(
            f"/search/v2/orgs/{self.org_id}/processes/search_jobs", method="POST", params=params
        )

    async def get_process_result_v2(
        self, job_id: str, start_row: Optional[str] = None, row_count: Optional[str] = None
    ) -> Dict:
        """
        This endpoint retrieves results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f"/search/v2/orgs/{self.org_id}/processes/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return await self._api_request(url)

    async def create_process_details_result_v2(self, process_guids: List[str], limited: bool = True) -> Dict:
        """
        Creates a process detail job. The results for the detail job may be requested using the job ID returned.
        This route will not request facets.
        :param process_guids: list of process guids
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v2/orgs/{self.org_id}/processes/detail_jobs"
        params = {"process_guids": process_guids}
        if limited:
            params["limited"] = limited

        return await self._api_request(url, method="POST", params=params)

    async def get_process_details_result_v2(self, job_id: str) -> Dict:
        """
        This endpoint retrieves results from the given query_id.
        :param job_id: Query ID to find results for
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v2/orgs/{self.org_id}/processes/detail_jobs/{job_id}/results"

        return await self._api_request(url)

    async def create_enriched_events_result_v2(
        self,
        query: str,
        criteria: Optional[Dict[str, Sequence[str]]] = None,
        exclusions: Optional[Dict[str, Sequence[str]]] = None,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        start: Optional[int] = None,
        time_range: Optional[Dict[str, str]] = None,
    ) -> Dict:
        """
        Create an enriched events search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param query: a lucene query like "process_name:svchost.exe"
        :param criteria: in the format {"device_name": ["Win7x64"]}
        :param exclusions: in the format {"device_name": ["Win7x64"]}
        :param fields: list of fields in the format ["process_username"]
        :param rows: integer
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param start: integer
        :param time_range: in the format {"start": "2020-01-18T18:34:04Z", "end": "2020-01-18T18:34:04Z"}
        :return: an ApiResponse with either a valid response or error information
        """
        params = self._fill_params(
            query=query,
            criteria=criteria,
            exclusions=exclusions,
            fields=fields,
            rows=rows,
            sort=sort,
            start=start,
            time_range=time_range,
        )
        return await self._api_request(
            f"/search/v2/orgs/{self.org_id}/enriched_events/search_jobs", method="POST", params=params
        )

    async def get_enriched_events_result_v2(
        self, job_id: str, start_row: Optional[str] = None, row_count: Optional[str] = None
    ) -> Dict:
        """
        This endpoint retrieves enriched results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f"/search/v2/orgs/{self.org_id}/enriched_events/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return await self._api_request(url)

    async def create_enriched_events_details_result_v2(self, event_ids: List[str]) -> Dict:
        """
        Create an enriched events detail job. The results for the detail job may be requested using the job ID returned.
        This route will not request facets.
        :param event_ids: list of event ids
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v2/orgs/{self.org_id}/enriched_events/detail_jobs"
        params = {"event_ids": event_ids}

        return await self._api_request(url, method="POST", params=params)

    async def get_enriched_events_details_result_v2(self, job_id: str) -> Dict:
        """
        This endpoint retrieves enriched results from the given query_id.
        :param job_id: Query ID to find results for
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v2/orgs/{self.org_id}/enriched_events/detail_jobs/{job_id}/results"

        return await self._api_request(url)

    async def create_process_summary_v2(self, process_guid: str, parent_guid: Optional[str]) -> Dict:
        """
        Creates a process summary job for a given process_guid (and parent_guid optionally).
        :param process_guid: the guid of the process
        :param parent_guid: (optional) the guid of the parent process
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"process_guid": process_guid}
        if parent_guid is not None:
            params.update(parent_guid=parent_guid)
        url = f"/search/v2/orgs/{self.org_id}/processes/summary_jobs"
        return await self._api_request(url, method="POST", params=params)

    async def get_process_summary_v2(self, job_id: str, tree: bool = False) -> Dict:
        """
        This endpoint returns a summary for a specific process.
        :param job_id: Query ID to find results for
        :param tree: False - summary format (default), True - tree format
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v2/orgs/{self.org_id}/processes/summary_jobs/{job_id}/results"
        if tree:
            url += "?format=tree"
        return await self._api_request(url)


class QueryFactory:
    @staticmethod
    def get_from_string(org_id: str, s: str):
        query = AsyncSearchAPI(org_id=org_id)
        route_string = s.lower()
        route_dict = {
            "start": query.create_process_result_v2,
            "results": query.get_process_result_v2,
            "start_details": query.create_process_details_result_v2,
            "results_details": query.get_process_details_result_v2,
            "start_enriched": query.create_enriched_events_result_v2,
            "results_enriched": query.get_enriched_events_result_v2,
            "start_enriched_details": query.create_enriched_events_details_result_v2,
            "results_enriched_details": query.get_enriched_events_details_result_v2,
            "start_summary": query.create_process_summary_v2,
            "results_summary": query.get_process_summary_v2,
        }
        try:
            return route_dict[route_string]
        except KeyError:
            # pylint: disable=consider-iterating-dictionary
            # pylint: disable=raise-missing-from
            raise InvalidQueryError(f"{s} is not recognized as a valid route. " f"Must be one of {route_dict.keys()}")
