import json
import logging
from asyncio import Queue, sleep
from random import choices
from typing import AnyStr, Dict, Sequence

from pscr_common.aws import get_boto3_client
from pscr_common.stream import stream

from query_tool.operations.configuration import MacroOperation, OperationConfigurationFile
from query_tool.operations.configuration import Mode
from query_tool.settings import get_settings
from query_tool.workers.abstract_worker import AbstractWorker

_logger = logging.getLogger(__name__)


class OrgSelectionStrategy:
    def __init__(self):
        self.settings = get_settings()
        self.mode = Mode[self.settings.execution_mode]
        self.bucket = self.settings.admin_bucket
        self.number_buckets = self.settings.number_buckets
        self.env_id = self.settings.env_id

    def _get_yaml_config_orgs(self) -> Dict[AnyStr, Dict[AnyStr, float]]:
        _logger.info("loading orgs from local yaml config file")
        non_prod_orgs = OperationConfigurationFile(self.settings.config_path).raw["env_config"]["non_prod_orgs"]
        try:
            if self.env_id == "dev":
                return non_prod_orgs["dev"]
            if self.env_id == "staging":
                return non_prod_orgs["staging"]
        except ValueError:
            _logger.error("wrong env input")

    def _acquire_s3_org_file(self) -> Dict[AnyStr, Dict[AnyStr, float]]:
        s3 = get_boto3_client("s3")
        obj = s3.get_object(Bucket=self.bucket, Key="org_mapper/config.json")
        payload = obj["Body"].read().decode("utf-8")
        return json.loads(payload)

    @staticmethod
    def _filter_for_default(orgs: Dict[AnyStr, Dict[AnyStr, float]]) -> Dict[AnyStr, Dict[AnyStr, float]]:
        #  We only care about entries with the default storage class.
        new_orgs = {}
        for item in orgs:
            org, org_class = item.split(":")
            if org_class == "default":
                new_orgs[org] = orgs[item]
        return new_orgs

    @staticmethod
    def _get_weighted_array(orgs: Dict[AnyStr, Dict[AnyStr, float]]) -> Sequence[str]:
        orgs = OrgSelectionStrategy._filter_for_default(orgs)
        arr = []
        tupl_arr = (
            stream(orgs.items())
            .map(lambda t: (t[0], t[1]["Value"]))
            .sort(key=lambda t: t[1], reverse=True)
            .take(len(orgs.items()) // 2)
            .to_list()
        )

        for org_id, value in tupl_arr:
            for _ in range(max(int(value // 1024), 1)):
                arr.append(org_id)
        return arr

    @staticmethod
    def _chunks(lst, number_of_buckets):
        arrangement = [0] * number_of_buckets
        buckets = [[] for _ in range(number_of_buckets)]
        for el_id in range(len(lst)):
            bucket_id = el_id % number_of_buckets
            arrangement[bucket_id] = arrangement[bucket_id] + 1
        for bucket_id in range(number_of_buckets):
            for _ in range(arrangement[bucket_id]):
                buckets[bucket_id].append(lst.pop(0))
        return buckets

    # pylint: disable=redefined-argument-from-local
    def _get_current_buckets_top(self, orgs: Dict[AnyStr, Dict[AnyStr, float]]) -> Sequence[str]:
        orgs = self._filter_for_default(orgs)
        arr = []
        for org_id, value in ((k, v["Value"]) for k, v in orgs.items()):
            if "-" not in org_id:
                arr.append((org_id, value))

        # This sorts all orgs per size, reverses it and gets it into a proper list of orgs
        reverse_sorted_org_list = stream(arr).sort(key=lambda t: t[1], reverse=True).map(lambda t: t[0]).to_list()
        # ...and chunks it into the number of buckets that we want resulting into a list of list of orgs.
        buckets = self._chunks(reverse_sorted_org_list, self.number_buckets)

        selected_orgs = []
        for orgs in buckets:
            selected_orgs.append(orgs[0])
        # Reversed, will start with smallest org so that 1 = small org, big number = large org
        return list(reversed(selected_orgs))

    def execute(self, test_mode: bool):
        if test_mode:
            return [str(i) for i in range(1, 10)]
        resp_orgs = []
        orgs = {}
        if self.env_id in ["dev", "staging"]:
            resp_orgs = self._get_yaml_config_orgs()
            _logger.info(
                f"Non-prod env: Acquired local file with {len(resp_orgs)} elements from . Sample: {list(resp_orgs)[:3]}"
            )
        else:
            orgs = self._acquire_s3_org_file()
            if self.mode in [
                Mode.availability,
                Mode.availability_summary,
                Mode.availability_details,
            ]:
                weighted_array = self._get_weighted_array(orgs)
                resp_orgs = choices(weighted_array, k=self.number_buckets) if weighted_array else []
            elif self.mode == Mode.performance:
                resp_orgs = self._get_current_buckets_top(orgs)
            _logger.info(f"Prod env: Acquired s3 file with {len(resp_orgs)} elements. Sample: {list(resp_orgs)[:3]}")

        return resp_orgs


class DriverWorker(AbstractWorker):
    def __init__(self, macro_op_queue: Queue):
        super().__init__(macro_op_queue)
        self.org_select = OrgSelectionStrategy()

    async def run(self):
        wait_time = get_settings().query_interval
        operations = OperationConfigurationFile(get_settings().config_path).operations
        _logger.info("Starting DriverWorker")
        while True:
            q_position = 1
            orgs = list(self.org_select.execute(self.test_mode))
            if not orgs:
                _logger.warning("No orgs selected!")
                _logger.debug(f"Waiting for {wait_time} seconds.")
                await sleep(wait_time)
                continue

            for org in orgs:
                for op in operations:
                    macro_op = MacroOperation(op, org, q_position)
                    _logger.debug(f"Adding macro op {macro_op} to queue.")
                    await self.macro_op_queue.put(macro_op)
                q_position += 1

                if self.test_mode:
                    break

                _logger.debug(f"{self.macro_op_queue.qsize()} elements in queue. Waiting for {wait_time} seconds.")
                await sleep(wait_time)
            if self.test_mode:
                break
