import logging
import os
import traceback
from asyncio import Queue, sleep, TimeoutError
from time import time

import requests
from aiohttp.client_exceptions import ClientConnectorError
from aio_pyformance.registry import global_registry
from aio_pyformance.reporters.reporter import Reporter
from pyformance.reporters.console_reporter import ConsoleReporter
from pscr_common.metrics import get_reporter
from pscr_common.metadata import MetadataProvider

from query_tool.settings import get_settings
from query_tool.workers.abstract_worker import AbstractWorker

_logger = logging.getLogger(__name__)


def get_container_id():
    metadata_uri = os.environ.get("ECS_CONTAINER_METADATA_URI_V4", None)
    if metadata_uri:
        _logger.info("Requesting Container Data")
        containers = requests.get(f"{metadata_uri}/task", timeout=600).json()["Containers"]
        _logger.debug(f"Containers: {containers}")
        short_container_id = containers[0]["DockerId"][0:12]
        # Assume we are the only container in the task.
        _logger.info(f"Container ID: {short_container_id}")
        return short_container_id
    return "unknown"


class MetricsWorker(AbstractWorker):
    def __init__(self, macro_op_queue: Queue):
        super().__init__(macro_op_queue)
        self.registry = global_registry()
        self.reporter_interval = None
        self.reporter = self._reporter_factory()

    def _reporter_factory(self) -> Reporter:
        settings = get_settings()
        self.reporter_interval = settings.reporter_interval

        env = os.environ.get("ENV_ID", "unknown")
        env_meta = MetadataProvider().get_metadata()
        reporter = get_reporter(
            self.registry,
            env,
            env_meta.task_id,
            "query-tool",
            env_meta.image.rsplit(":", maxsplit=1)[-1],
        )

        if not reporter:
            reporter = ConsoleReporter(registry=self.registry, reporting_interval=self.reporter_interval)

        return reporter

    async def run(self):
        await super().run()
        _logger.info(f'Starting MetricsWorker of type "{get_settings().metrics_type.lower()}"')
        while True:
            _logger.debug("Reporting metrics batch")
            try:
                await self.reporter.report_now(timestamp=int(time()))
            except TimeoutError:
                _logger.error(f"Timeout reporting metrics:\n{traceback.format_exc()}")
            except ClientConnectorError:
                _logger.error(f"Failed to report metrics:\n{traceback.format_exc()}")
            if self.test_mode and self.macro_op_queue.empty():
                break
            # Trigger a coroutine switch
            await sleep(self.reporter_interval)

    def dispose(self):
        super().dispose()
